@@ -7,7 +7,7 @@ use Carp;
 
 use base 'DynaLoader';
 
-our $VERSION = '0.09';
+our $VERSION = '0.10';
 bootstrap RRD::Tweak;
 
 # Internal object structure:
@@ -249,8 +249,8 @@ sub validate {
                 return 0;
             }
 
-            if( $r->{$key} !~ /^-?nan$/i and
-                $r->{$key} !~ /^[0-9e+\-.]+$/i ) {
+            if( $r->{$key} !~ /^-?nan$/io and
+                $r->{$key} !~ /^[0-9e+\-.]+$/io ) {
                 $self->_set_errmsg('$self->{ds}[' . $ds .
                                    ']{' . $key . '} is not a number');
                 return 0;
@@ -992,7 +992,7 @@ sub modify_ds {
                 }
 
                 if( $val =~ /^-?nan$/i ) {
-                    if( $ds_attr->{$key} !~ /^-?nan$/i ) {
+                    if( $ds_attr->{$key} !~ /^-?nan$/io ) {
                         $ds_attr->{$key} = 'nan';
                     }
                 }
@@ -1121,113 +1121,8 @@ sub add_rra {
 
     push(@{$self->{'cdp_prep'}}, $rra_cdp_prep);
 
-    # try to derive the data rows where possible ($n_rra holds the old
-    # number of arrays)
-    if( $n_rra > 0 ) {
-
-        my %rraidx_per_steps;
-        for( my $rra=0; $rra < $n_rra; $rra++) {
-            if( $self->{'rra'}[$rra]->{'cf'} eq $cf ) {
-                my $steps = $self->{'rra'}[$rra]->{'pdp_per_row'};
-                $rraidx_per_steps{$steps} = $rra;
-            }
-        }
-
-        if( scalar(keys %rraidx_per_steps) > 0 ) {
-
-            # arrange the RRA indexes from most granular to less granular
-            my @rraidx_ascending_steps;
-            foreach my $steps (sort {$a <=> $b} keys %rraidx_per_steps) {
-                push( @rraidx_ascending_steps, $rraidx_per_steps{$steps} );
-            }
-
-            for( my $row=0; $row < $rra_len; $row++ ) {
-
-                # negative integers in (pdp_step)*seconds
-                my $row_start_time = ($row - $rra_len) * $pdp_per_row;
-                my $row_end_time = $row_start_time + $pdp_per_row;
-
-                foreach my $rra ( @rraidx_ascending_steps ) {
-                    my $src_steps = $self->{'rra'}[$rra]->{'pdp_per_row'};
-                    my $src_data = $self->{'cdp_data'}[$rra];
-                    my $src_len = scalar(@{$src_data});
-
-                    if( $row_start_time + $src_len * $src_steps < 0 ) {
-                        # our new row is outside of the array boundaries
-                        next;
-                    }
-
-                    my $n_src_rows = int($pdp_per_row/$src_steps);
-                    if( $n_src_rows == 0 ) {
-                        # the source row is less granular than ours --
-                        # then we take just one value
-                        $n_src_rows = 1;
-                    }
-
-                    # grab the values from source rows
-                    my @known_values = ();
-                    my @unknown_values = ();
-                    my $start_src_row = int($src_len +
-                                            $row_start_time/$src_steps);
-                    for( my $ds=0; $ds < $n_ds; $ds++ ) {
-
-                        my $known_val_per_ds = [];
-                        my $unknown_val_per_ds = 0;
-
-                        for( my $src_row_pos = 0; $src_row_pos < $n_src_rows;
-                             $src_row_pos++ ) {
-                            my $src_row = $start_src_row + $src_row_pos;
-                            my $data_element = $src_data->[$src_row][$ds];
-                            if( $data_element =~ /nan/io ) {
-                                $unknown_val_per_ds++;
-                            }
-                            else {
-                                push(@{$known_val_per_ds}, $data_element);
-                            }
-                        }
-
-                        push( @known_values, $known_val_per_ds );
-                        push( @unknown_values, $unknown_val_per_ds );
-                    }
-
-                    # if the number of knowns is good enough, take the value
-                    for( my $ds=0; $ds < $n_ds; $ds++ ) {
-                        if( $unknown_values[$ds] * 1.0 / $n_src_rows <
-                            $arg->{xff} ) {
-
-                            # now calculate the new value from knowns
-
-                            if( $cf eq 'AVERAGE' ) {
-                                my $val = 0;
-                                map {$val += $_} @{$known_values[$ds]};
-                                $rra_data->[$row][$ds] =
-                                    $val / scalar(@{$known_values[$ds]});
-                            }
-                            elsif( $cf eq 'MIN' ) {
-                                my $val = '+inf';
-                                map {if($_ < $val){$val = $_} }
-                                    @{$known_values[$ds]};
-                                $rra_data->[$row][$ds] = $val;
-                            }
-                            elsif( $cf eq 'MAX' ) {
-                                my $val = '-inf';
-                                map {if($_ > $val){$val = $_} }
-                                    @{$known_values[$ds]};
-                                $rra_data->[$row][$ds] = $val;
-                            }
-                            elsif( $cf eq 'LAST' ) {
-                                my $val = pop @{$known_values[$ds]};
-                                $rra_data->[$row][$ds] = $val;
-                            }
-                        }
-                    }
-
-                    # we've done all we could, so finish trying other RRA's
-                    last;
-                }
-            }
-        }
-    }
+    # try to derive the data rows where possible
+    $self->_populate_rra($n_rra);
 
     return;
 }
@@ -1357,6 +1252,9 @@ sub modify_rra {
             }
 
             unshift(@{$rra_data}, @{$prepend_rra_data});
+            
+            # try to derive the data rows where possible
+            $self->_populate_rra($mod_rra_index);
         }
     }
     return;
@@ -1458,7 +1356,7 @@ sub has_hwpredict {
     }
     return 0;
 }
-    
+
 
 
 
@@ -1490,7 +1388,7 @@ sub ds_descr {
         $ret .= ':' . $ds_attr->{'hb'};
 
         foreach my $key ('min', 'max') {
-            if( $ds_attr->{$key} !~ /^-?nan$/i ) {
+            if( $ds_attr->{$key} !~ /^-?nan$/io ) {
                 $ret .= ':' . $ds_attr->{$key};
             }
             else {
@@ -1547,8 +1445,151 @@ sub rra_descr {
 }
 
 
+##
+##  Private methods
+###################
 
+# after an RRA is created or extended, this method tries to derive the
+# values from other RRA's
 
+sub _populate_rra
+{
+    my $self = shift;
+    my $pop_rra_index = shift;
+
+    my $rra_data = $self->{'cdp_data'}[$pop_rra_index];
+    my $rra_len = scalar(@{$rra_data});
+    my $cf = $self->{'rra'}[$pop_rra_index]{'cf'};
+    my $pdp_per_row = $self->{'rra'}[$pop_rra_index]{'pdp_per_row'};
+    my $xff = $self->{'rra'}[$pop_rra_index]{'xff'};
+
+    my $n_rra = scalar(@{$self->{'rra'}});
+    my $n_ds = scalar(@{$self->{'ds'}});
+
+    if( $n_rra == 1 ) {
+        # this is the only RRA, so there's no data to derive from
+        return;
+    }
+
+    my %rraidx_per_steps;
+    for( my $rra=0; $rra < $n_rra; $rra++) {
+        if( $rra != $pop_rra_index ) {
+            my $r = $self->{'rra'}[$rra];
+            my $steps = $r->{'pdp_per_row'};
+
+            if( ($r->{'cf'} eq $cf) or
+                ($r->{'cf'} eq 'AVERAGE' and $steps == 1) ) {
+                $rraidx_per_steps{$steps} = $rra;
+            }
+        }
+    }
+
+    if( scalar(keys %rraidx_per_steps) == 0 ) {
+        # we could not find any RRA to derive the data from
+        return;
+    }
+
+    # arrange the RRA indexes from most granular to less granular
+    my @rraidx_ascending_steps;
+    foreach my $steps (sort {$a <=> $b} keys %rraidx_per_steps) {
+        push( @rraidx_ascending_steps, $rraidx_per_steps{$steps} );
+    }
+
+    for ( my $row=0; $row < $rra_len; $row++ ) {
+
+        # check if we have any NANs in this row
+        my @nan_ds;
+        for ( my $ds=0; $ds < $n_ds; $ds++ ) {
+            if( $rra_data->[$row][$ds] =~ /nan/io ) {
+                push(@nan_ds, $ds);
+            }
+        }
+
+        if( not scalar(@nan_ds) ) {
+            # this row has all values already defined, skip it from population
+            next;
+        }
+
+        # negative integers in (pdp_step)*seconds
+        my $row_start_time = ($row - $rra_len) * $pdp_per_row;
+        my $row_end_time = $row_start_time + $pdp_per_row;
+
+        foreach my $rra ( @rraidx_ascending_steps ) {
+            my $r = $self->{'rra'}[$rra];
+            my $src_steps = $r->{'pdp_per_row'};
+            my $src_data = $self->{'cdp_data'}[$rra];
+            my $src_len = scalar(@{$src_data});
+
+            if ( $row_start_time + $src_len * $src_steps < 0 ) {
+                # our new row is outside of the array boundaries
+                next;
+            }
+
+            my $n_src_rows = int($pdp_per_row/$src_steps);
+            if ( $n_src_rows == 0 ) {
+                # the source row is less granular than ours --
+                # then we take just one value
+                $n_src_rows = 1;
+            }
+
+            # grab the values from source rows
+            my @known_values = ();
+            my @unknown_values = ();
+            my $start_src_row = int($src_len +
+                                    $row_start_time/$src_steps);
+            foreach my $ds (@nan_ds) {
+                my $known_val_per_ds = [];
+                my $unknown_val_per_ds = 0;
+
+                for ( my $src_row_pos = 0; $src_row_pos < $n_src_rows;
+                      $src_row_pos++ ) {
+                    my $src_row = $start_src_row + $src_row_pos;
+                    my $data_element = $src_data->[$src_row][$ds];
+                    if ( $data_element =~ /nan/io ) {
+                        $unknown_val_per_ds++;
+                    } else {
+                        push(@{$known_val_per_ds}, $data_element);
+                    }
+                }
+
+                $known_values[$ds] = $known_val_per_ds;
+                $unknown_values[$ds] = $unknown_val_per_ds;
+            }
+
+            # if the number of knowns is good enough, take the value
+            foreach my $ds (@nan_ds) {
+                if ( $unknown_values[$ds] * 1.0 / $n_src_rows <= $xff ) {
+
+                    # now calculate the new value from knowns
+                    if ( $cf eq 'AVERAGE' ) {
+                        my $val = 0;
+                        map {$val += $_} @{$known_values[$ds]};
+                        $rra_data->[$row][$ds] =
+                            $val / scalar(@{$known_values[$ds]});
+                    } elsif ( $cf eq 'MIN' ) {
+                        my $val = '+inf';
+                        map {if($_ < $val){$val = $_} }
+                            @{$known_values[$ds]};
+                        $rra_data->[$row][$ds] = $val;
+                    } elsif ( $cf eq 'MAX' ) {
+                        my $val = '-inf';
+                        map {if($_ > $val){$val = $_} }
+                            @{$known_values[$ds]};
+                        $rra_data->[$row][$ds] = $val;
+                    } elsif ( $cf eq 'LAST' ) {
+                        my $val = pop @{$known_values[$ds]};
+                        $rra_data->[$row][$ds] = $val;
+                    }
+                }
+            }
+
+            # we've done all we could, so finish trying other RRA's
+            last;
+        }
+    }
+
+    return;
+}
 
 
 
